@@ -177,10 +177,10 @@ Rmpfr_fmodquo Rmpfr_fpif_export Rmpfr_fpif_import Rmpfr_flags_clear Rmpfr_flags_
 Rmpfr_flags_test Rmpfr_flags_save Rmpfr_flags_restore Rmpfr_rint_roundeven Rmpfr_roundeven
 Rmpfr_nrandom Rmpfr_erandom Rmpfr_fmma Rmpfr_fmms Rmpfr_log_ui Rmpfr_gamma_inc Rmpfr_beta
 Rmpfr_round_nearest_away rndna
-atonv nvtoa atodouble Rmpfr_dot Rmpfr_get_str_ndigits
+atonv nvtoa atodouble doubletoa Rmpfr_dot Rmpfr_get_str_ndigits
 );
 
-    our $VERSION = '4.09';
+    our $VERSION = '4.12';
     #$VERSION = eval $VERSION;
 
     DynaLoader::bootstrap Math::MPFR $VERSION;
@@ -282,7 +282,7 @@ Rmpfr_fmodquo Rmpfr_fpif_export Rmpfr_fpif_import Rmpfr_flags_clear Rmpfr_flags_
 Rmpfr_flags_test Rmpfr_flags_save Rmpfr_flags_restore Rmpfr_rint_roundeven Rmpfr_roundeven
 Rmpfr_nrandom Rmpfr_erandom Rmpfr_fmma Rmpfr_fmms Rmpfr_log_ui Rmpfr_gamma_inc Rmpfr_beta
 Rmpfr_round_nearest_away rndna
-atonv nvtoa atodouble Rmpfr_dot Rmpfr_get_str_ndigits
+atonv nvtoa atodouble doubletoa Rmpfr_dot Rmpfr_get_str_ndigits
 )]);
 
 
@@ -292,6 +292,11 @@ $Math::MPFR::NNW = 0; # Set to 1 to allow "non-numeric" warnings for operations
 $Math::MPFR::NOK_POK = 0; # Set to 1 to allow warnings in new() and overloaded operations when
                           # a scalar that has set both NOK (NV) and POK (PV) flags is encountered
 
+$Math::MPFR::doubletoa_fallback = 0; # If FALLBACK_NOTIFY is defined, this scalar Will be automatically
+                                     # incremented whenever the grisu3 algorithm (used by doubletoa) fails
+                                     # to produce correct result, and thus falls back to its designated
+                                     # fallback routine. (See the doubletoa documentation for details.)
+
 %Math::MPFR::NV_properties = _get_NV_properties();
 
 sub dl_load_flags {0} # Prevent DynaLoader from complaining and croaking
@@ -794,81 +799,8 @@ sub _get_NV_properties {
   return %properties;
 }
 
-
-sub nvtoa {
-
-   if(defined $Math::MPFR::NV_properties{type}) { die "$Math::MPFR::NV_properties{type}" }
-   my $nv = shift;
-   my $significand_sign = '';
-   if($nv <= 0) {
-     if($nv == 0) {
-       if($] >= 5.010 && scalar(reverse(unpack('h*', pack('F<', $nv)))) =~ /8/   ) {
-         $significand_sign = '-';
-       }
-     }
-     else {
-       $significand_sign = '-';
-       $nv *= -1;
-     }
-   }
-
-   my @s = _nvtoa($nv, $Math::MPFR::NV_properties{NV_MAX}, $Math::MPFR::NV_properties{normal_min},
-                $Math::MPFR::NV_properties{min_pow},$Math::MPFR::NV_properties{bits},
-                $Math::MPFR::NV_properties{max_dig});
-
-   return "$significand_sign$s[0]" if @s == 1;
-
-
-   my $leading_zero_count = 0;
-   my $exp_sign = '';
-   my $output = $s[0];
-   my $k = $s[1];
-
-   for(0 .. length($output)) {
-     if(substr($output, $_, 1) eq '0') {
-       $leading_zero_count++;
-       $k--;
-     }
-     else { last }
-   }
-
-   return $significand_sign . _std($output, $k);
-}
-
-# standardize the string returned by the _nvtoa() XSub.
-sub _std {
-  my $output = shift;
-  my $k = shift;
-
-  my $len  = length($output);
-
-  my $critical = $k + $len;
-
-  if($critical < -3) {
-
-    $critical--;
-    $critical = sprintf "%03d", $critical;
-    if($len > 1) { substr($output, 1, 0, '.') }
-    return "${output}e${critical}";
-  }
-
-  if($critical <= 0 && $critical >= -3) { return $output = "0." . ('0' x -$critical) . $output }
-  if($critical > 0 && $critical < $Math::MPFR::NV_properties{max_dig}) {
-
-    if($k >= 0) {
-      $output .= '0' x $k;
-      return $output . ".0";
-    }
-
-    substr($output, $k, 0, '.');
-    return $output;
-  }
-
-    # $critical >= $max_dig
-
-    $critical--;
-    if($len > 1) { substr($output, 1, 0, '.') }
-    return "${output}e+${critical}";
+sub perl_set_fallback_flag {
+  $Math::MPFR::doubletoa_fallback++;
 }
 
 *Rmpfr_get_z_exp             = \&Rmpfr_get_z_2exp;
